from ..basetypes import BaseResponse, Command, Mode, Protocol, Response
from ..protocol import BaseProtocol


class ProtocolCAN(BaseProtocol):
    """Supported Protocols:
    - [0x06] ISO 15765-4 CAN (11 bit ID, 500 Kbaud)
    - [0x07] ISO 15765-4 CAN (29 bit ID, 500 Kbaud)
    - [0x08] ISO 15765-4 CAN (11 bit ID, 250 Kbaud)
    - [0x09] ISO 15765-4 CAN (29 bit ID, 250 Kbaud)
    - [0x0A] SAE J1939 CAN (29 bit ID, 250 Kbaud)
    - [0x0B] USER1 CAN (11 bit ID, 125 Kbaud)
    - [0x0C] USER2 CAN (11 bit ID, 50 Kbaud)
    """
    def parse_response(self, base_response: BaseResponse, command: Command) -> Response:
        if command.mode == Mode.AT: # AT Commands
            status = None
            if len(base_response.message[:-1]) == 1:
                status = ''.join([c.decode(errors="ignore") for c in base_response.message[0]]).strip()

            return Response(**base_response.__dict__, value=status)
        else: # OBD Commands
            value = None
            parsed_data = list()
            for raw_line in base_response.message[:-1]: # Skip the last line (prompt character)
                line = ''.join([c.decode(errors="ignore") for c in raw_line]).strip()
                
                components = line.split(' ')

                if len(components) <= 4: # Skip header-less lines
                    continue

                header = components[0]
                bytes_offset = 2 # Mode and PID offset
                length = int(components[1], 16) - bytes_offset

                data = components[-length:]

                parsed_data.append(data)
            
            if command.formula:
                try:
                    value = command.formula(parsed_data)
                except Exception:
                    value = None

            return Response(**base_response.__dict__, parsed_data=parsed_data, value=value)


ProtocolCAN.register(
    Protocol.ISO_15765_4_CAN, Protocol.ISO_15765_4_CAN_B, Protocol.ISO_15765_4_CAN_C,
    Protocol.ISO_15765_4_CAN_D, Protocol.SAE_J1939_CAN, Protocol.USER1_CAN, Protocol.USER2_CAN
)


from functools import partial

from ..basetypes import BaseMode, Command, Mode

M = Mode.AT
C = partial(Command, M, min_value=None, max_value=None, units=None)

# ELM327.pdf | AT Command Summary | Page 8 - 9

class ModeAT(BaseMode):
    """AT Commands"""

    # General Commands
    REPEAT = Command(Mode.NONE, '', 0x00, "REPEAT", "Repeat the last command")
    BAUDRATE_DIVISOR = C("BRD {hh}", 0x00, "BAUDRATE_DIVISOR", "Try Baude Rate Divisor hh", command_args={"hh": int})
    SET_BAUDRATE_TIMEOUT = C("BRT {hh}", 0x00, "SET_BAUDRATE_TIMEOUT", "Set Baud Rate Timeout", command_args={"hh": int})
    TO_DEFAULT = C('D', 0x00, "TO_DEFAULT", "Set all to defaults")
    ECHO_OFF = C("E0", 0x00, "ECHO_OFF", "Echo off")
    ECHO_ON = C("E1", 0x00, "ECHO_ON", "Echo on")
    FORGET_EVENTS = C("FE", 0x00, "FORGET_EVENTS", "Forget Events")
    VERSION_ID = C('I', 0x00, "VERSION_ID", "Print the version ID")
    LINEFEED_OFF = C("L0", 0x00, "LINEFEED_OFF", "Linefeeds off")
    LINEFEED_ON = C("L1", 0x00, "LINEFEED_ON", "Linefeeds on")
    LOWPOWER = C("LP", 0x00, "LOWPOWER", "Go to Low Power mode")
    MEMORY_OFF = C("M0", 0x00, "MEMORY_OFF", "Memory off")
    MEMORY_ON = C("M0", 0x00, "MEMORY_ON", "Memory on")
    READ_STORED = C("RD", 0x00, "READ_STORED", "Read the stored Data")
    SAVE_DATA = C("SD {hh}", 0x00, "SAVE_DATA", "Save Data byte hh", command_args={"hh": int})
    SOFT_RESET = C("WS", 0x00, "SOFT_RESET", "Warm Start (quick software reset)")
    RESET = C('Z', 0x00, "RESET", "Reset all")
    DESCRIPTION = C("@1", 0x00, "DESCRIPTION", "Display device description")
    IDENTIFIER = C("@2", 0x00, "IDENTIFIER", "Display device identifier")
    STORE_ID = C("@3 {cccccccccccc}", 0x00, "STORE_ID", "Stores the @2 device identifier", command_args={"cccccccccccc": int})

    # Programmable Parameter Commands
    PROG_PARAM_OFF = C("PP {xx} OFF", 0x00, "PROG_PARAM_OFF", "Disable Prog Parameter xx", command_args={"xx": str})
    PROG_PARAMS_OFF = C("PP FF OFF", 0x00, "PROG_PARAMS_OFF", "All Prog Parameters off")
    PROG_PARAM_ON = C("PP {xx} ON", 0x00, "PROG_PARAM_ON", "Enable Prog Parameter xx", command_args={"xx": str})
    PROG_PARAMS_ON = C("PP FF ON", 0x00, "PROG_ON", "All Prog Parameters on")
    PROG_SET_PARAM = C("PP {xx} SV {yy}", 0x00, "PROG_SET_PARAM", "For PP xx, Set the Value to yy", command_args={"xx": str, "yy": str})
    PROG_SUMMARY = C("PPS", 0x00, "PROG_SUMMARY", "Print a PP Summary")

    # Voltage Reading Commands
    CALIBRATE_VOLTAGE = C("CV {dddd}", 0x00, "CALIBRATE_VOLTAGE", "Calibrate the Voltage to dd.dd volts", command_args={"dddd": int})
    RESTORE_VOLTAGE = C("CV 0000", 0x00, "RESTORE_VOLTAGE", "Restore CV value to factory setting")
    READ_VOLTAGE = C("RV", 0x00, "READ_VOLTAGE", "Read the Voltage")

    # Other
    IGN = C("IGN", 0x00, "IGN", "read the IgnMon input level")

    # OBD Command
    ALLOW_LONG = C("AL", 0x00, "ALLOW_LONG", "Allow Long (>7 bytes) messages")
    AUTO_RECEIVE = C("AR", 0x00, "AUTO_RECEIVE", "Automatically Receive")
    ADAP_TIMING_OFF = C("AT0", 0x00, "ADAP_TIMING_OFF", "Adaptive Timing off")
    ADAP_TIMING_1 = C("AT1", 0x00, "ADAP_TIMING_1", "Adaptive Timing Auto1")
    ADAP_TIMING_2 = C("AT2", 0x00, "ADAP_TIMING_2", "Adaptive Timing Auto2")
    BUFFER_DUMP = C("BD", 0x00, "BUFFER_DUMP", "Perform a Buffer Dump")
    BYPASS_INIT = C("BI", 0x00, "BYPASS_INIT", "Bypass the Initialization sequence")
    DESC_PROTOCOL = C("DP", 0x00, "DESC_PROTOCOL", "Describe the current Protocol")
    DESC_PROTOCOL_N = C("DPN", 0x00, "DESC_PROTOCOL_N", "Describe the Protocol by Number")
    HEADERS_OFF = C("H0", 0x00, "HEADERS_OFF", "Headers off")
    HEADERS_ON = C("H1", 0x00, "HEADERS_ON", "Headers on")
    MONITOR_ALL = C("MA", 0x00, "MONITOR_ALL", "Monitor All")
    MONITOR_RECEIVER = C("MR {hh}", 0x00, "MONITOR_RECEIVER", "Monitor for Receiver = hh", command_args={"hh": int})
    MONITOR_TRANSMITTER = C("MT {hh}", 0x00, "MONITOR_TRANSMITTER", "Monitor for Transmitter = hh", command_args={"hh": int})
    NORMAL_LENGTH = C("NL", 0x00, "NORMAL_LENGTH", "Normal Length messages")
    PROTOCOL_CLOSE = C("PC", 0x00, "PROTOCOL_CLOSE", "Protocol Close")
    RESPONSES_OFF = C("R0", 0x00, "RESPONSES_OFF", "Responses off")
    RESPONSES_ON = C("R1", 0x00, "RESPONSES_ON", "Responses on")
    SET_RECEIVE_ADDR = C("RA {hh}", 0x00, "SET_RECEIVE_ADDR", "Set Receive Address to hh", command_args={"hh": int})
    SPACES_OFF = C("S0", 0x00, "SPACES_OFF", "Printing of Spaces off")
    SPACES_ON = C("S1", 0x00, "SPACES_ON", "Printing of Spaces on")
    SET_HEADER = C("SH {x}{y}{z}", 0x00, "SET_HEADER", "Set Header to xyz", command_args={'x': str, 'y': str, 'z': str})
    SET_HEADER_LONG = C("SH {xx}{yy}{zz}", 0x00, "SET_HEADER_LONG", "Set Header to xxyyzz", command_args={"xx": str, "yy": str, "zz": str})
    SET_PROTOCOL = C("SP {h}", 0x00, "SET_PROTOCOL", "Set Protocol to h and save it", command_args={"h": int})
    SET_PROTOCOL_AUTO = C("SP A{h}", 0x00, "SET_PROTOCOL_AUTO", "Set Protocol to Auto, h and save it", command_args={"h": int})
    SET_RECEIVE_ADDR_ALT = C("SR {hh}", 0x00, "SET_RECEIVE_ADDR_ALT", "Set the receive address to hh", command_args={"hh": int})
    STANDARD_SEARCH = C("SS", 0x00, "STANDARD_SEARCH", "Use Standard Search order (J1978)")
    SET_TIMEOUT = C("ST {hh}", 0x00, "SET_TIMEOUT", "Set Timeout to hh x 4 msec", command_args={"hh": int})
    SET_TESTER_ADDR = C("TA {hh}", 0x00, "SET_TESTER_ADDR", "Set Tester Address to hh", command_args={"hh": int})
    TRY_PROTOCOL = C("TP {h}", 0x00, "TRY_PROTOCOL", "Try Protocol h", command_args={"h": int})
    TRY_PROTOCOL_AUTO = C("TP A{h}", 0x00, "TRY_PROTOCOL_AUTO", "Try Protocol h with Auto search", command_args={"h": int})

    # ISO Specific Commands
    FAST_INIT = C("FI", 0x00, "FAST_INIT", "Perform a Fast Initiation")
    SET_BAUDRATE_10400 = C("IB 10", 0x00, "SET_BAUDRATE_10400", "Set the ISO Baud rate to 10400")
    SET_BAUDRATE_4800 = C("IB 48", 0x00, "SET_BAUDRATE_4800", "Set the ISO Baud rate to 4800")
    SET_BAUDRATE_9600 = C("IB 96", 0x00, "SET_BAUDRATE_9600", "Set the ISO Baud rate to 9600")
    SET_ISO_INIT_ADDR = C("IIA {hh}", 0x00, "SET_ISO_INIT_ADDR", "Set ISO (slow) Init Address to hh", command_args={"hh": int})
    KEY_WORDS = C("KW", 0x00, "KEY_WORDS", "Display the Key Words")
    KEY_WORD_OFF = C("KW0", 0x00, "KEY_WORD_OFF", "Key Word checking off")
    KEY_WORD_ON = C("KW1", 0x00, "KEY_WORD_ON", "Key Word checking on")
    SLOW_INIT = C("SI", 0x00, "SLO_INIT", "Perform a Slow (5 baud) Initiation")
    SET_WAKEUP = C("SW {hh}", 0x00, "SET_WAKEUP", "Set Wakeup interval to hh x 20 msec", command_args={"hh": int})
    # WAKEUP_MESSAGE = C("", 0x00, "WAKEUP_MESSAGE", "Set the Wakeup Message")

    # J1850 Specific Commands
    IFR_OFF = C("IFR0", 0x00, "IFR_OFF", "IFRs off")
    IFR_AUTO = C("IFR1", 0x00, "IFR_AUTO", "IFRs Auto")
    IFR_ON = C("IFR2", 0x00, "IFR_ON", "IFRs on")
    IFR_HEADER = C("IFR H", 0x00, "IFR_HEADER", "IFR value from Header")
    IFR_SOURCE = C("IFR S", 0x00, "IFR_SOURCE", "IFR value from Source")
    
    # CAN Specific Commands
    CAN_EXT_ADDR_OFF = C("CEA", 0x00, "CAN_EXT_ADDR_OFF", "Turn off CAN Extended Addressing")
    CAN_EXT_ADDR = C("CEA {hh}", 0x00, "CAN_EXT_ADDR", "Use CAN Extended Address hh", command_args={"hh": int})
    FORMAT_AUTO_OFF = C("CAF0", 0x00, "FORMAT_AUTO_OFF", "Automatic Formatting off")
    FORMAT_AUTO_ON = C("CAF0", 0x00, "FORMAT_AUTO_ON", "Automatic Formatting on")
    SET_ID_FILTER = C("CF {hhh}", 0x00, "SET_ID_FILTER", "Set the ID Filter to hhh", command_args={"hhh": int})
    SET_ID_FILTER_LONG = C("CF {hhhhhhhh}", 0x00, "SET_ID_FILTER_LONG", "Set the ID Filter to hhhhhhhh", command_args={"hhhhhhhh": int})
    FLOW_CONTROLS_OFF = C("CFC0", 0x00, "FLOW_CONTROLS_OFF", "Flow Controls off")
    FLOW_CONTROLS_ON = C("CFC1", 0x00, "FLOW_CONTROLS_ON", "Flow Controls on")
    SET_ID_MASK = C("CM {hhh}", 0x00, "SET_ID_MASK", "Set the ID Mask to hhh", command_args={"hhh": int})
    SET_ID_MASK_LONG = C("CM {hhhhhhhh}", 0x00, "SET_ID_MASK_LONG", "Set the ID Mask to hhhhhhhh", command_args={"hhhhhhhh": int})
    SET_CAN_PRIORITY = C("CP {hh}", 0x00, "SET_CAN_PRIORITY", "Set CAN Priority to hh (29 bit)", command_args={"hh": int})
    SET_CAN_ADDR = C("CRA {hhh}", 0x00, "SET_CAN_ADDR", "Set CAN Receive Address to hhh", command_args={"hhh": int})
    SET_CAN_ADDR_LONG = C("CRA {hhhhhhhh}", 0x00, "SET_CAN_ADDR_LONG", "Set CAN Receive Address to hhhhhhhh", command_args={"hhhhhhhh": int})
    CAN_STATUS = C("CS", 0x00, "CAN_STATUS", "Show the CAN Status counts")
    DLC_OFF = C("D0", 0x00, "DLC_OFF", "Display of the DLC off")
    DLC_ON = C("D1", 0x00, "DLC_ON", "Display of the DLC on")
    SET_FLOW_CONTROL_MODE = C("FC SM {h}", 0x00, "SET_FLOW_CONTROL_MODE", "Flow Control, Set the Mode to h", command_args={"h": int})
    SET_FLOW_CONTROL_HEADER = C("FC SH {hhh}", 0x00, "SET_FLOW_CONTROL_HEADER", "Flow Control, Set the Header to hhh", command_args={"hhh": int})
    SET_FLOW_CONTROL_HEADER_LONG = C("FC SH {hhhhhhhh}", 0x00, "SET_FLOW_CONTROL_HEADER_LONG", "Flow Control, Set the Header to hhhhhhhh", command_args={"hhhhhhhh": int})
    # SET_FLOW_CONTROL_DATA = C("", 0x00, "SET_FLOW_CONTROL_DATA", "Flow Control, Set Data to [...]")
    PROTOCOL_B_BAUDRATE = C("PB {xx} {yy}", 0x00, "PROTOCOL_B_BAUDRATE", "Protocol B options and baud rate", command_args={"xx": str, "yy": str})
    RTR_MESSAGE = C("RTR", 0x00, "RTR_MESSAGE", "Send an RTR message")
    VARIABLE_DLC_OFF = C("V0", 0x00, "VARIABLE_DLC_OFF", "Use of Variable DLC off")
    VARIABLE_DLC_ON = C("V1", 0x00, "VARIABLE_DLC_OFF", "Use of Variable DLC on")
    
    # J1939 CAN Specific Commands
    MONITOR_DM1 = C("DM1", 0x00, "MONITOR_DM1", "Monitor for DM1 messages")
    FORMAT_ELM = C("JE", 0x00, "FORMAT_ELM", "Use J1939 Elm data format")
    FORMAT_SAE = C("JS", 0x00, "FORMAT_SAE", "Use J1939 SAE data format")
    MONITOR_PGN = C("MP {hhhh}", 0x00, "MONITOR_PGN", "Monitor for PGN 0hhhh", command_args={"hhhh": int})
    MONITOR_PGN_LONG = C("MP {hhhhhh}", 0x00, "MONITOR_PGN_LONG", "Monitor for PGN hhhhhh", command_args={"hhhhhh": int})

# Initialize AT Commands
at_commands = ModeAT()


from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Type, Union


from .basetypes import BaseResponse, Command, Protocol, Response


class BaseProtocol(ABC):
    _registry: Dict[Protocol, Type["BaseProtocol"]] = {}

    extra_init_sequence: List[Union[Command, Callable]]

    def __init__(self) -> None: ...

    @abstractmethod
    def parse_response(self, base_response: BaseResponse, command: Command) -> Response: ...

    @classmethod
    def register(cls, *protocols: Protocol) -> None:
        """Register a subclass with its supported protocols."""
        for protocol in protocols:
            cls._registry[protocol] = cls
    
    @classmethod
    def get_handler(cls, protocol: Protocol) -> "BaseProtocol":
        """Retrieve the appropriate protocol class or fallback to ProtocolUnknown."""
        return cls._registry.get(protocol, ProtocolUnknown)()


class ProtocolJ1850(BaseProtocol):
    """Supported Protocols:
    - [0x01] SAE J1850 PWM (41.6 Kbaud)
    - [0x02] SAE J1850 VPW (10.4 Kbaud)
    """
    def parse_response(self, base_response: BaseResponse, command: Command) -> Response:
        raise NotImplementedError

ProtocolJ1850.register(Protocol.SAE_J1850_PWM, Protocol.SAE_J1850_VPW)


class ProtocolKWP2000(BaseProtocol):
    """Supported Protocols:
    - [0x03] ISO 9141-2 (5 baud init, 10.4 Kbaud)
    - [0x04] ISO 14230-4 KWP (5 baud init, 10.4 Kbaud)
    - [0x05] ISO 14230-4 KWP (fast init, 10.4 Kbaud)
    """
    def parse_response(self, base_response: BaseResponse, command: Command) -> Response:
        raise NotImplementedError

ProtocolKWP2000.register(Protocol.ISO_9141_2, Protocol.ISO_14230_4_KWP, Protocol.ISO_14230_4_KWP_FAST)


class ProtocolUnknown(BaseProtocol): 
    """Fallback protocol class for unknown or unsupported protocols.

    In such cases, basic serial communication might still be possible,
    but full message parsing could be limited.
    """
    def parse_response(self, base_response: BaseResponse, command: Command) -> Response:
        raise NotImplementedError


from typing import Union, overload

from .basetypes import BaseMode, Command

from .modes import Mode01

T_Modes = Union[Mode01]

class Modes(
    Mode01, 
    ):
    pass

class Commands(Modes):
    def __init__(self):
        self.modes = {
            0x01: Mode01(),
        }

    @overload
    def __getitem__(self, key: str) -> Command: ...

    @overload
    def __getitem__(self, key: int) -> T_Modes: ...

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, str):
            key = key.upper()
            if not key in dir(self):
                raise KeyError(f"Command '{key}' not found")
            item = getattr(self, key)
            if not isinstance(item, Command):
                raise TypeError(f"Expected Command but got {type(item)} for key '{key}'")
            return item
        elif isinstance(key, int):
            if not key in self.modes:
                raise KeyError(f"Mode '{key}' not found")
            mode = self.modes.get(key)
            if not isinstance(mode, BaseMode):
                raise TypeError(f"Expected Mode but got {type(mode)} for key '{key}'")
            return mode
        else:
            raise TypeError(f"Unsupported {type(key)} type")

# Initialize Commands
commands = Commands()


from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from time import time
from typing import Any, Callable, Dict, List, Optional, Union


class Mode(Enum):
    NONE = ''
    """Special mode used for the REPEAT command"""

    AT = "AT"
    """Special mode to send AT commands"""

    REQUEST = 0x01
    """Request current data"""
    FREEZE_FRAME = 0x02
    """Request freeze frame data"""
    STATUS_DTC = 0x03
    """Request stored DTCs (Diagnostic Trouble Codes)"""
    CLEAR_DTC = 0x04
    """Clear/reset DTCs (Diagnostic Trouble Codes)"""
    O2_SENSOR = 0x05
    """Request oxygen sensor monitoring test results"""
    PENDING_DTC = 0x06
    """Request DTCs (Diagnostic Trouble Codes) pending"""
    CONTROL_MODULE = 0x07
    """Request control module information"""
    O2_SENSOR_TEST = 0x08
    """Request oxygen sensor test results"""
    VEHICLE_INFO = 0x09
    """Request vehicle information"""
    PERMANENT_DTC = 0x0A
    """Request permanent DTCs (Diagnostic Trouble Codes)"""


class Protocol(Enum):
    UNKNOWN = -1
    """Unknown protocol"""

    AUTO = 0x00
    """Automatically determine the protocol"""
    SAE_J1850_PWM = 0x01
    """SAE J1850 PWM (41.6 kbaud)"""
    SAE_J1850_VPW = 0x02
    """SAE J1850 VPW (10.4 kbaud)"""
    ISO_9141_2 = 0x03
    """ISO 9141-2 (5 baud init, 10.4 kbaud)"""
    ISO_14230_4_KWP = 0x04
    """ISO 14230-4 KWP (5 baud init, 10.4 kbaud)"""
    ISO_14230_4_KWP_FAST = 0x05
    """ISO 14230-4 KWP (fast init, 10.4 kbaud)"""
    ISO_15765_4_CAN = 0x06
    """ISO 15765-4 CAN (11 bit ID, 500 kbaud)"""
    ISO_15765_4_CAN_B = 0x07
    """ISO 15765-4 CAN (29 bit ID, 500 kbaud)"""
    ISO_15765_4_CAN_C = 0x08
    """ISO 15765-4 CAN (11 bit ID, 250 kbaud)"""
    ISO_15765_4_CAN_D = 0x09
    """ISO 15765-4 CAN (29 bit ID, 250 kbaud)"""
    SAE_J1939_CAN = 0x0A
    """SAE J1939 CAN (29 bit ID, 250* kbaud) *default settings (user adjustable)"""
    USER1_CAN = 0x0B
    """USER1 CAN (11* bit ID, 125* kbaud) *default settings (user adjustable)"""
    USER2_CAN = 0x0C
    """USER2 CAN (11* bit ID, 50* kbaud) *default settings (user adjustable)"""

class Command():
    def __init__(self, 
            mode: Mode,
            pid: Union[int, str],
            n_bytes: int,
            name: str,
            description: Optional[str] = None,
            min_value: Optional[Union[int, float, str]] = None,
            max_value: Optional[Union[int, float, str]] = None,
            units: Optional[str] = None,
            formula: Optional[Callable] = None,
            command_args: Optional[Dict[str, Any]] = None,
        ) -> None:
        self.mode = mode
        self.pid = pid
        self.n_bytes = n_bytes
        self.name = name
        self.description = description
        self.min_value = min_value
        self.max_value = max_value
        self.units = units
        self.formula = formula

        self.command_args = command_args or {}
        self.is_formatted = False

    def __call__(self, *args: Any, checks: bool = True) -> "Command":
        if not self.command_args or not args or len(self.command_args) != len(args):
            raise TypeError(f"{self.__repr__()} expects {len(self.command_args)} argument(s), but got {len(args)}")
        
        fmt_command = deepcopy(self) # we don't want to modify the original command pid
        try:
            combined_args = {}
            for (arg, arg_type), value in zip(self.command_args.items(), args):
                if checks:
                    if not isinstance(value, arg_type):
                        raise TypeError(f"Argument '{arg}' should be of type {arg_type}")

                    if isinstance(value, int):
                        value = f"{value:0{len(arg)}X}"
                    elif isinstance(value, str) and len(value) != len(arg):
                        raise ValueError(f"Argument '{arg}' should have length {len(arg)}, but got {len(value)}")

                combined_args[arg] = value

            fmt_command.is_formatted = True
            fmt_command.pid = str(self.pid).format(**combined_args)
        except Exception as e:
            raise e
        return fmt_command

    def __repr__(self) -> str:
        return f"<Command {self.mode} {self.pid if isinstance(self.pid, str) else f'{self.pid:02X}'} {self.name or 'Unnamed'} [{', '.join(self.command_args.keys())}]>"
    
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Command):
            return False

        return (
            self.mode == value.mode and
            self.name == value.name and
            self.description == value.description and
            self.command_args == value.command_args
        )

    def build(self) -> bytes:
        """Builds the query to be sent.
        ELM327 is not case-sensitive, ignores spaces and all control characters."""
        if self.command_args and not self.is_formatted:
            raise ValueError(f"Command has unset arguments for '{self.pid}': {self.command_args}")

        mode = self.mode.value
        pid = self.pid
        if isinstance(mode, int):
            mode = f"{mode:02X}"
        if isinstance(pid, int):
            pid = f"{pid:02X}"
        return f"{mode} {pid}\r".encode() # \r ends each command


class BaseMode():
    def __getitem__(self, key) -> Command:
        if isinstance(key, int):
            for attr_name in dir(self):
                attr = getattr(self, attr_name)
                if hasattr(attr, "pid") and attr.pid == key:
                    return attr
        raise KeyError(f"No command found with PID {key}")
    
    def __repr__(self) -> str:
        return f"<Mode Commands: {len(self)}>"

    def __len__(self) -> int:
        return len([1 for attr_name in dir(self) if isinstance(getattr(self, attr_name), Command)])
    
    def has_command(self, command: Union[Command, str]) -> bool:
        if isinstance(command, Command):
            command = command.name

        command = command.upper()
        return hasattr(self, command) and isinstance(getattr(self, command), Command)


@dataclass
class BaseResponse():
    command: Command
    raw_response: List[bytes]
    message: List[List[bytes]]
    timestamp: float = field(default_factory=time)

@dataclass
class Response(BaseResponse):
    parsed_data: Optional[List[List[str]]] = None

    value: Optional[Any] = None

    @property
    def min_value(self) -> Optional[Union[int, float, str]]:
        return self.command.min_value
    
    @property
    def max_value(self) -> Optional[Union[int, float, str]]:
        return self.command.max_value
    
    @property
    def units(self) -> Optional[str]:
        return self.command.units


from functools import partial
from logging import Handler, Formatter, DEBUG, INFO, WARNING, ERROR, CRITICAL, StreamHandler, getLogger
from os import environ
from sys import platform
from typing import Any, List, Optional

from .basetypes import BaseResponse


def bytes_to_string(raw_response: List[bytes], filter_bytes: List[bytes] = []) -> str:
    filtered_response = [c for c in raw_response if c not in filter_bytes]
    return b''.join(filtered_response).decode(errors="ignore").strip()

filter_bts = partial(bytes_to_string, filter_bytes=[b'\r', b'>'])

def debug_baseresponse(base_response: BaseResponse) -> str:
    out = ''

    for line in base_response.message[:-1]: # omit prompt line
        out += f"[{filter_bts(line)}]\n"
    
    return out


def setup_logging(
    handler: Optional[Handler] = None,
    formatter: Optional[Formatter] = None,
    level: Optional[int] = None,
    root: bool = True,
) -> None:
    """A helper function to setup logging."""
    if not level:
        level = INFO

    if not handler:
        handler = StreamHandler()

    if not formatter:
        if isinstance(handler, StreamHandler) and _stream_supports_colour(handler.stream):
            formatter = _ColorFormatter()
        else:
            dt_fmt = "%Y-%m-%d %H:%M:%S"
            formatter = Formatter("[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style='{')

    if root:
        logger = getLogger()
    else:
        library, _, _ = __name__.partition('.')
        logger = getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)

def _stream_supports_colour(stream: Any) -> bool:
    is_a_tty = hasattr(stream, "isatty") and stream.isatty()

    if "PYCHARM_HOSTED" in environ or environ.get("TERM_PROGRAM") == "vscode":
        return is_a_tty

    if platform != "win32":
        return is_a_tty

    return is_a_tty and "WT_SESSION" in environ

class _ColorFormatter(Formatter):
    LEVEL_COLORS = [
        (DEBUG, "\x1b[40;1m"),
        (INFO, "\x1b[34;1m"),
        (WARNING, "\x1b[33;1m"),
        (ERROR, "\x1b[31m"),
        (CRITICAL, "\x1b[41m"),
    ]

    FORMATS = {
        level: Formatter(
            f"\x1b[30;1m%(asctime)s\x1b[0m {colour}%(levelname)-8s\x1b[0m \x1b[35m%(name)s\x1b[0m %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
        for level, colour in LEVEL_COLORS
    }

    def format(self, record):
        formatter = self.FORMATS.get(record.levelno)
        if formatter is None:
            formatter = self.FORMATS[DEBUG]

        if record.exc_info:
            text = formatter.formatException(record.exc_info)
            record.exc_text = f"\x1b[31m{text}\x1b[0m"

        output = formatter.format(record)

        record.exc_text = None
        return output



from logging import Formatter, Handler, getLogger, INFO
from re import IGNORECASE, search as research
from serial import Serial, SerialException # type: ignore
from typing import Callable, List, Optional, Union

from .utils import debug_baseresponse, filter_bts, setup_logging

from .basetypes import BaseResponse, Command, Protocol, Response
from .modes import ModeAT
from .protocol import BaseProtocol

_log = getLogger(__name__)


class Connection():
    def __init__(self, 
                    port: str,
                    baudrate: int = 38400,
                    protocol: Protocol = Protocol.AUTO,
                    timeout: float = 5.0,
                    write_timeout: float = 3.0,
                    auto_connect: bool = True,
                    smart_query: bool = False,
                    *,
                    log_handler: Optional[Handler] = None,
                    log_formatter: Optional[Formatter] = None,
                    log_level: Optional[int] = INFO,
                    log_root: bool = False,
                ) -> None:
        """Initialize connection settings and auto-connect by default.

        Parameters
        -----------
        port: :class:`str`
            The serial port (e.g., "COM5", "/dev/ttyUSB0", "/dev/rfcomm0").
        baudrate: :class:`int`
            The baud rate for communication (e.g., 38400, 115200).
        protocol: :class:`Protocol`
            The protocol to use for communication (default: Protocol.AUTO).
        timeout: :class:`float`
            The maximum time (in seconds) to wait for a response from the device before raising a timeout error (default: 5.0).
        write_timeout: :class:`float`
            The time (in seconds) to wait for data to be written to the device before raising a timeout error (default: 3.0).
        auto_connect: Optional[:class:`bool`]
            If set to true, method connect will be called.
        smart_query: Optional[:class:`bool`]
            If set to true, and if the same command is sent twice, the second time it will be sent as a repeat command.

        *:
            Keyword-only arguments (non-positional).


        log_handler: Optional[:class:`logging.Handler`]
            The log handler to use for the library's logger.
        log_formatter: :class:`logging.Formatter`
            The formatter to use with the given log handler.
        log_level: :class:`int`
            The default log level for the library's logger.
        root_logger: :class:`bool`
            Whether to set up the root logger rather than the library logger.
        """
        self.port = port
        self.baudrate = baudrate
        self.protocol = protocol
        self.timeout = timeout
        self.write_timeout = write_timeout
        self.smart_query = smart_query

        self.serial_conn: Optional[Serial] = None
        self.protocol_handler = BaseProtocol.get_handler(Protocol.UNKNOWN)
        self.supported_protocols: List[Protocol] = []
        self.last_command: Optional[Command] = None

        self.init_sequence: List[Union[Command, Callable[[], None]]] = [
            ModeAT.RESET,
            ModeAT.ECHO_OFF,
            ModeAT.LINEFEED_OFF,
            ModeAT.HEADERS_ON,
            ModeAT.SPACES_ON,
            self._auto_protocol,
        ]

        self.protocol_preferences = [
            Protocol.ISO_15765_4_CAN,       # 0x06
            Protocol.ISO_15765_4_CAN_B,     # 0x07
            Protocol.ISO_15765_4_CAN_C,     # 0x08
            Protocol.ISO_15765_4_CAN_D,     # 0x09
            Protocol.SAE_J1850_PWM,         # 0x01
            Protocol.SAE_J1850_VPW,         # 0x02
            Protocol.ISO_9141_2,            # 0x03
            Protocol.ISO_14230_4_KWP,       # 0x04
            Protocol.ISO_14230_4_KWP_FAST,  # 0x05 
            Protocol.SAE_J1939_CAN,         # 0x0A
            Protocol.USER1_CAN,             # 0x0B
            Protocol.USER2_CAN,             # 0x0C
        ]

        if log_handler or log_formatter or log_level:
            setup_logging(
                log_handler,
                log_formatter,
                log_level,
                log_root
            )

        if auto_connect:
            self.connect()

    def connect(self, **kwargs) -> None:
        """Establishes a connection and initializes the device."""
        try:
            _log.info(f"Attempting to connect to {self.port} at {self.baudrate} baud.")
            self.serial_conn = Serial(
                port=self.port, 
                baudrate=self.baudrate, 
                timeout=self.timeout, 
                write_timeout=self.write_timeout,
                **kwargs
            )
            self._initialize_connection()
            _log.info(f"Successfully connected to {self.port}.")
        except SerialException as e:
            self.serial_conn = None
            _log.error(f"Failed to connect to {self.port}: {e}")
            raise ConnectionError(f"Failed to connect: {e}")
        
    def _initialize_connection(self) -> None:
        """Initializes the connection using the init sequence."""
        for command in self.init_sequence:
            if isinstance(command, Command):
                self.query(command)
            elif callable(command):
                command()
            else:
                _log.error(f"Invalid type in init_sequence: {type(command)}")
                raise TypeError(f"Invalid command type: {type(command)}")

    def _auto_protocol(self, protocol: Optional[Protocol] = None) -> None:
        """Sets the protocol for communication."""
        protocol = protocol or self.protocol

        protocol_number = self._set_protocol_to(protocol)

        if protocol_number in [0, -1]:
            self.supported_protocols = self._get_supported_protocols()
            supported_protocols = self.supported_protocols

            if supported_protocols:
                priority_dict = {protocol: idx for idx, protocol in enumerate(self.protocol_preferences)}
                supported_protocols.sort(key=lambda p: priority_dict.get(p, len(self.protocol_preferences)))

                protocol_number = self._set_protocol_to(supported_protocols[0])
            else:
                protocol_number = -1

        self.protocol = Protocol(protocol_number)
        self.protocol_handler = BaseProtocol.get_handler(self.protocol)
        _log.info(f"Protocol set to {self.protocol.name}.")

    def _set_protocol_to(self, protocol: Protocol) -> int:
        """Attempts to set the protocol to the specified value, return the protocol number if successful."""
        self.query(ModeAT.SET_PROTOCOL(protocol.value))
        response = self.query(ModeAT.DESC_PROTOCOL_N)

        line = filter_bts(response.raw_response)
        protocol_number = self._parse_protocol_number(line)

        return protocol_number

    def _get_supported_protocols(self) -> List[Protocol]:
        """Attempts to find supported protocol(s)."""
        supported_protocols = []

        for protocol in Protocol:
            if protocol in [Protocol.UNKNOWN, Protocol.AUTO]:
                continue

            protocol_number = self._set_protocol_to(protocol)
            if protocol_number == protocol.value:
                supported_protocols.append(protocol)
        
        if not supported_protocols:
            _log.warning("No supported protocols detected.")
            supported_protocols = [Protocol.UNKNOWN]

        return supported_protocols

    def _parse_protocol_number(self, line: str) -> int:
        """Extracts and returns the protocol number from the response line."""
        match = research(r"([0-9A-F])$", line, IGNORECASE)
        if match:
            return int(match.group(1), 16)
        return -1

    def _send_query(self, query: bytes) -> None:
        """Sends a query to the ELM327."""
        if not self.serial_conn or not self.serial_conn.is_open:
            _log.error("Attempted to send a query without an active connection.")
            raise ConnectionError("Attempted to send a query without an active connection.")

        self.clear_buffer()
        self.serial_conn.write(query)
        self.serial_conn.flush()
    
    def _read_byte(self) -> bytes:
        if not self.serial_conn or not self.serial_conn.is_open:
            _log.error("Attempted to read without an active connection.")
            raise ConnectionError("Attempted to read without an active connection.")
        
        return self.serial_conn.read(1)

    def query(self, command: Command) -> Response:
        """Sends a command and waits for a response."""        
        if self.smart_query and self.last_command and command == self.last_command:
            query = ModeAT.REPEAT.build()
        else:
            query = command.build()

        _log.debug(f">>> Send: {str(query)}")

        self._send_query(query)
        self.last_command = command

        return self.wait_for_response(command)

    def wait_for_response(self, command: Command) -> Response:
        """Reads data dynamically until the OBDII prompt (>) or timeout."""
        raw_response: List[bytes] = []

        message: List[List[bytes]] = []
        current_line: List[bytes] = []
        while True:
            chunk = self._read_byte()
            if not chunk: # Timeout
                break
            raw_response.append(chunk)
            char = chunk.decode(errors="ignore")

            if char in ['\r', '\n']:
                if current_line:
                    message.append(current_line)
                    current_line = []
                continue
            current_line.append(chunk)
            if char == '>': # Ending prompt character
                break
        if current_line:
            message.append(current_line)

        base_response = BaseResponse(command, raw_response, message)

        _log.debug(f"<<< Read:\n{debug_baseresponse(base_response)}")

        try:
            return self.protocol_handler.parse_response(base_response, command)
        except NotImplementedError:
            return Response(**base_response.__dict__)
    
    def clear_buffer(self) -> None:
        """Clears any buffered input from the adapter."""
        if self.serial_conn:
            self.serial_conn.reset_input_buffer()

    def close(self) -> None:
        """Close the serial connection if not already done."""
        if self.serial_conn:
            self.serial_conn.close()
            _log.debug("Connection closed.")


import pytest

from ast import parse

from obdii.formula import Formula, SafeEvaluator


@pytest.mark.parametrize(
    "expression, variables, expected_result",
    [
        ("a + b", {"a": 1, "b": 2}, 3),
        ("a - b", {"a": 3, "b": 1}, 2),
        ("a * b", {"a": 2, "b": 3}, 6),
        ("a / b", {"a": 6, "b": 2}, 3),
        ("a // b", {"a": 7, "b": 3}, 2),
        ("a % b", {"a": 7, "b": 3}, 1),
        ("a ** b", {"a": 2, "b": 3}, 8),
        ("a ^ b", {"a": 5, "b": 3}, 6),
    ]
)
def test_safe_evaluator(expression, variables, expected_result):
    evaluator = SafeEvaluator(variables)

    parsed_expr = parse(expression, mode="eval")
    result = evaluator.visit(parsed_expr.body)

    assert result == expected_result


@pytest.mark.parametrize(
    "expression, parsed_data, expected_result",
    [
        ("a + b", [["1", "2"]], 3),
        ("a - b", [["3", "1"]], 2),
        ("a * b", [["2", "3"]], 6),
        ("a / b", [["6", "2"]], 3),

        ("a + b", [["1", "A"]], 11),
        ("a - b", [["A", "5"]], 5),
        ("a * b", [["A", "3"]], 30),
        ("a / b", [["A", "2"]], 5),

        ("a + b", [["1", "AB"]], 172),
        ("a - b", [["AB", "9"]], 162),
        ("a * b", [["AB", "4"]], 684),
        ("a / b", [["AC", "2"]], 86),
    ]
)
def test_formula(expression, parsed_data, expected_result):
    formula = Formula(expression)

    result = formula(parsed_data)

    assert result == expected_result


@pytest.mark.parametrize(
    "expression, parsed_data, expected_result",
    [
        ("(a + b) * c", [["1", "2", "3"]], 9),
        ("a + (b * c)", [["1", "2", "3"]], 7),
        ("(a + b) - (c / d)", [["1", "2", "3", "1"]], 0),
        ("a + (b - c) * d", [["6", "4", "2", "3"]], 12),
        ("(a - b) * (c + d)", [["8", "4", "2", "2"]], 16),
        ("(a + (b * c)) - d", [["1", "2", "3", "4"]], 3),
        ("(a * (b + c)) - d", [["2", "3", "4", "1"]], 13),
        ("(a + b) * (c + (d - e))", [["1", "2", "3", "4", "5"]], 6),
    ]
)
def test_formula_with_parentheses(expression, parsed_data, expected_result):
    formula = Formula(expression)

    result = formula(parsed_data)

    assert result == expected_result


def test_formula_invalid_input():
    formula = Formula("a + b")

    with pytest.raises(ValueError):
        formula([])



from functools import partial

from ..basetypes import BaseMode, Command, Mode
from ..formula import Formula

M = Mode.REQUEST
C = partial(Command, M)

F = Formula

# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_01_-_Show_current_data

class Mode01(BaseMode):
    """Request Commands"""

    PIDS_A = C(0x00, 0x04, "PIDS_A", "PIDs supported [$01 - $20]", None, None, None)
    STATUS = C(0x01, 0x04, "STATUS", "Monitor status since DTCs cleared. (Includes malfunction indicator lamp (MIL), status and number of DTCs, components tests, DTC readiness checks)", None, None, None)
    FREEZE_DTC = C(0x02, 0x02, "FREEZE_DTC", "DTC that caused freeze frame to be stored.", None, None, None)
    FUEL_STATUS = C(0x03, 0x02, "FUEL_STATUS", "Fuel system status", None, None, None)
    ENGINE_LOAD = C(0x04, 0x01, "ENGINE_LOAD", "Calculated engine load", 0, 100, '%', F("100/255*A"))
    COOLANT_TEMP = C(0x05, 0x01, "COOLANT_TEMP", "Engine coolant temperature", -40, 215, "°C", F("A-40"))
    STFT_B1 = C(0x06, 0x01, "STFT_B1", "Short term fuel trim (STFT)—Bank 1", -100, 99.2, '%', F("100/128*A-100"))
    LTFT_B1 = C(0x07, 0x01, "LTFT_B1", "Long term fuel trim (LTFT)—Bank 1", -100, 99.2, '%', F("100/128*A-100"))
    STFT_B2 = C(0x08, 0x01, "STFT_B2", "Short term fuel trim (STFT)—Bank 2", -100, 99.2, '%', F("100/128*A-100"))
    LTFT_B2 = C(0x09, 0x01, "LTFT_B2", "Long term fuel trim (LTFT)—Bank 2", -100, 99.2, '%', F("100/128*A-100"))
    FUEL_PRESSURE = C(0x0A, 0x01, "FUEL_PRESSURE", "Fuel pressure (gauge pressure)", 0, 765, "kPa", F("3*A"))
    INTAKE_PRESSURE = C(0x0B, 0x01, "INTAKE_PRESSURE", "Intake manifold absolute pressure", 0, 255, "kPa", F('A'))
    ENGINE_SPEED = C(0x0C, 0x02, "ENGINE_SPEED", "Engine speed", 0, 16383.75, "rpm", F("(256*A+B)/4"))
    VEHICLE_SPEED = C(0x0D, 0x01, "VEHICLE_SPEED", "Vehicle speed", 0, 255, "km/h", F('A'))
    TIMING_ADVANCE = C(0x0E, 0x01, "TIMING_ADVANCE", "Timing advance", -64, 63.5, "° before TDC", F("A/2-64"))
    INTAKE_TEMP = C(0x0F, 0x01, "INTAKE_TEMP", "Intake air temperature", -40, 215, "°C", F("A-40"))
    MAF = C(0x10, 0x02, "MAF", "Mass air flow sensor (MAF) air flow rate", 0, 655.35, "g/s", F("(256*A+B)/100"))
    THROTTLE_POSITION = C(0x11, 0x01, "THROTTLE_POSITION", "Throttle position", 0, 100, '%', F("100/255*A"))
    SEC_AIR_STATUS = C(0x12, 0x01, "SEC_AIR_STATUS", "Commanded secondary air status", None, None, None)
    O2_SENSORS = C(0x13, 0x01, "O2_SENSORS", "Oxygen sensors present (in 2 banks)", None, None, None)
    O2_S1 = C(0x14, 0x02, "O2_S1", "Oxygen Sensor 1 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S2 = C(0x15, 0x02, "O2_S2", "Oxygen Sensor 2 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S3 = C(0x16, 0x02, "O2_S3", "Oxygen Sensor 3 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S4 = C(0x17, 0x02, "O2_S4", "Oxygen Sensor 4 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S5 = C(0x18, 0x02, "O2_S5", "Oxygen Sensor 5 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S6 = C(0x19, 0x02, "O2_S6", "Oxygen Sensor 6 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S7 = C(0x1A, 0x02, "O2_S7", "Oxygen Sensor 7 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    O2_S8 = C(0x1B, 0x02, "O2_S8", "Oxygen Sensor 8 A: Voltage B: Short term fuel trim", "0V -100%", "1.275V 99.2%", "V %")
    OBD_STANDARDS = C(0x1C, 0x01, "OBD_STANDARDS", "OBD standards this vehicle conforms to", 1, 250, None)
    O2_SENSORS_ALT = C(0x1D, 0x01, "O2_SENSORS_ALT", "Oxygen sensors present (in 4 banks)", None, None, None)
    AUX_INPUT_STATUS = C(0x1E, 0x01, "AUX_INPUT_STATUS", "Auxiliary input status", None, None, None)
    RUN_TIME = C(0x1F, 0x02, "RUN_TIME", "Run time since engine start", 0, 65535, 's', F("256*A+B"))

    PIDS_B = C(0x20, 0x04, "PIDS_B", "PIDs supported [$21 - $40]", None, None, None)
    DISTANCE_W_MIL = C(0x21, 0x02, "DISTANCE_W_MIL", "Distance traveled with malfunction indicator lamp (MIL) on", 0, 65535, "km", F("256*A+B"))
    FUEL_RAIL_PRESSURE_VAC = C(0x22, 0x02, "FUEL_RAIL_PRESSURE_VAC", "Fuel Rail Pressure (relative to manifold vacuum)", 0, 5177.265, "kPa", F("0.079*(256*A+B)"))
    FUEL_RAIL_GAUGE_PRESSURE = C(0x23, 0x02, "FUEL_RAIL_GAUGE_PRESSURE", "Fuel Rail Gauge Pressure (diesel, or gasoline direct injection)", 0, 655350, "kPa", F("10*(256*A+B)"))
    O2_S1_LAMBDA = C(0x24, 0x04, "O2_S1_LAMBDA", "Oxygen Sensor 1 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S2_LAMBDA = C(0x25, 0x04, "O2_S2_LAMBDA", "Oxygen Sensor 2 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S3_LAMBDA = C(0x26, 0x04, "O2_S3_LAMBDA", "Oxygen Sensor 3 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S4_LAMBDA = C(0x27, 0x04, "O2_S4_LAMBDA", "Oxygen Sensor 4 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S5_LAMBDA = C(0x28, 0x04, "O2_S5_LAMBDA", "Oxygen Sensor 5 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S6_LAMBDA = C(0x29, 0x04, "O2_S6_LAMBDA", "Oxygen Sensor 6 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S7_LAMBDA = C(0x2A, 0x04, "O2_S7_LAMBDA", "Oxygen Sensor 7 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    O2_S8_LAMBDA = C(0x2B, 0x04, "O2_S8_LAMBDA", "Oxygen Sensor 8 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Voltage", "0ratio 0V", "<2ratio <8V", "ratio V")
    COMMANDED_EGR = C(0x2C, 0x01, "COMMANDED_EGR", "Commanded EGR", 0, 100, '%', F("100/255*A"))
    EGR_ERROR = C(0x2D, 0x01, "EGR_ERROR", "EGR Error", -100, 99.2, '%', F("100/128*A-100"))
    EVAPORATIVE_PURGE = C(0x2E, 0x01, "EVAPORATIVE_PURGE", "Commanded evaporative purge", 0, 100, '%', F("100/255*A"))
    FUEL_LEVEL = C(0x2F, 0x01, "FUEL_LEVEL", "Fuel Tank Level Input", 0, 100, '%', F("100/255*A"))
    WARMUPS_SINCE_DTC_CLEARED = C(0x30, 0x01, "WARMUPS_SINCE_DTC_CLEARED", "Warm-ups since codes cleared", 0, 255, None, F('A'))
    DISTANCE_SINCE_DTC_CLEARED = C(0x31, 0x02, "DISTANCE_SINCE_DTC_CLEARED", "Distance traveled since codes cleared", 0, 65535, "km", F("256*A+B"))
    EVAP_VAPOR_PRESSURE = C(0x32, 0x02, "EVAP_VAPOR_PRESSURE", "Evap. System Vapor Pressure", -8192, 8191.75, "Pa")
    BAROMETRIC_PRESSURE = C(0x33, 0x01, "BAROMETRIC_PRESSURE", "Absolute Barometric Pressure", 0, 255, "kPa", F('A'))
    O2_S1_CURRENT = C(0x34, 0x04, "O2_S1_CURRENT", "Oxygen Sensor 1 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S2_CURRENT = C(0x35, 0x04, "O2_S2_CURRENT", "Oxygen Sensor 2 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S3_CURRENT = C(0x36, 0x04, "O2_S3_CURRENT", "Oxygen Sensor 3 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S4_CURRENT = C(0x37, 0x04, "O2_S4_CURRENT", "Oxygen Sensor 4 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S5_CURRENT = C(0x38, 0x04, "O2_S5_CURRENT", "Oxygen Sensor 5 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S6_CURRENT = C(0x39, 0x04, "O2_S6_CURRENT", "Oxygen Sensor 6 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S7_CURRENT = C(0x3A, 0x04, "O2_S7_CURRENT", "Oxygen Sensor 7 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    O2_S8_CURRENT = C(0x3B, 0x04, "O2_S8_CURRENT", "Oxygen Sensor 8 AB: Air-Fuel Equivalence Ratio (lambda,λ) CD: Current", "0ratio -128mA", "<2ratio <128mA", "ratio mA")
    CATALYST_TEMP_B1_S1 = C(0x3C, 0x02, "CATALYST_TEMP_B1_S1", "Catalyst Temperature: Bank 1, Sensor 1", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_B2_S1 = C(0x3D, 0x02, "CATALYST_TEMP_B2_S1", "Catalyst Temperature: Bank 2, Sensor 1", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_B1_S2 = C(0x3E, 0x02, "CATALYST_TEMP_B1_S2", "Catalyst Temperature: Bank 1, Sensor 2", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_B2_S2 = C(0x3F, 0x02, "CATALYST_TEMP_B2_S2", "Catalyst Temperature: Bank 2, Sensor 2", -40, 6513.5, "°C", F("(256*A+B)/10-40"))

    PIDS_C = C(0x40, 0x04, "PIDS_C", "PIDs supported [$41 - $60]", None, None, None)
    STATUS_DRIVE_CYCLE = C(0x41, 0x04, "STATUS_DRIVE_CYCLE", "Monitor status this drive cycle", None, None, None)
    CONTROL_MODULE_VOLTAGE = C(0x42, 0x02, "CONTROL_MODULE_VOLTAGE", "Control module voltage", 0, 65.535, 'V', F("(256*A+B)/1000"))
    ABSOLUTE_LOAD = C(0x43, 0x02, "ABSOLUTE_LOAD", "Absolute load value", 0, 25700, '%', F("100/255*(256*A+B)"))
    COMMANDED_AIR_FUEL_EQUIV_RATIO = C(0x44, 0x02, "COMMANDED_AIR_FUEL_EQUIV_RATIO", "Commanded Air-Fuel Equivalence Ratio (lambda,λ)", '0', "<2", "ratio", F("2/65536*(256*A+B)"))
    RELATIVE_THROTTLE_POSITION = C(0x45, 0x01, "RELATIVE_THROTTLE_POSITION", "Relative throttle position", 0, 100, '%', F("100/255*A"))
    AMBIENT_AIR_TEMP = C(0x46, 0x01, "AMBIENT_AIR_TEMP", "Ambient air temperature", -40, 215, "°C", F("A-40"))
    THROTTLE_POSITION_B = C(0x47, 0x01, "THROTTLE_POSITION_B", "Absolute throttle position B", 0, 100, '%', F("100/255*A"))
    THROTTLE_POSITION_C = C(0x48, 0x01, "THROTTLE_POSITION_C", "Absolute throttle position C", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_D = C(0x49, 0x01, "ACCELERATOR_POSITION_D", "Accelerator pedal position D", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_E = C(0x4A, 0x01, "ACCELERATOR_POSITION_E", "Accelerator pedal position E", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_F = C(0x4B, 0x01, "ACCELERATOR_POSITION_F", "Accelerator pedal position F", 0, 100, '%', F("100/255*A"))
    THROTTLE_ACTUATOR = C(0x4C, 0x01, "THROTTLE_ACTUATOR", "Commanded throttle actuator", 0, 100, '%', F("100/255*A"))
    RUN_TIME_W_MIL = C(0x4D, 0x02, "RUN_TIME_W_MIL", "Time run with MIL on", 0, 65535, "min", F("256*A+B"))
    TIME_SINCE_DTC_CLEARED = C(0x4E, 0x02, "TIME_SINCE_DTC_CLEARED", "Time since trouble codes cleared", 0, 65535, "min", F("256*A+B"))
    MAX_FUEL_AIR_RATIO_O2_VOLT_CURR_PRESSURE = C(0x4F, 0x04, "MAX_FUEL_AIR_RATIO_O2_VOLT_CURR_PRESSURE", "Maximum value for Fuel-Air equivalence ratio, oxygen sensor voltage, oxygen sensor current, and intake manifold absolute pressure", "0, 0, 0, 0", "255, 255, 255, 2550", "ratio, V, mA, kPa")
    MAX_MAF = C(0x50, 0x04, "MAX_MAF", "Maximum value for air flow rate from mass air flow sensor", 0, 2550, "g/s")
    FUEL_TYPE = C(0x51, 0x01, "FUEL_TYPE", "Fuel Type", None, None, None)
    ETHANOL_PERC = C(0x52, 0x01, "ETHANOL_PERC", "Ethanol fuel %", 0, 100, '%', F("100/255*A"))
    EVAP_VAPOR_PRESSURE_ABS = C(0x53, 0x02, "EVAP_VAPOR_PRESSURE_ABS", "Absolute Evap system Vapor Pressure", 0, 327.675, "kPa", F("(256*A+B)/200"))
    EVAP_VAPOR_PRESSURE_ALT = C(0x54, 0x02, "EVAP_VAPOR_PRESSURE_ALT", "Evap system vapor pressure", -32768, 32767, "Pa")
    SHORT_O2_TRIM_B1 = C(0x55, 0x02, "SHORT_O2_TRIM_B1", "Short term secondary oxygen sensor trim, A: bank 1, B: bank 3", -100, 99.2, '%')
    LONG_O2_TRIM_B1 = C(0x56, 0x02, "LONG_O2_TRIM_B1", "Long term secondary oxygen sensor trim, A: bank 1, B: bank 3", -100, 99.2, '%')
    SHORT_O2_TRIM_B2 = C(0x57, 0x02, "SHORT_O2_TRIM_B2", "Short term secondary oxygen sensor trim, A: bank 2, B: bank 4", -100, 99.2, '%')
    LONG_O2_TRIM_B2 = C(0x58, 0x02, "LONG_O2_TRIM_B2", "Long term secondary oxygen sensor trim, A: bank 2, B: bank 4", -100, 99.2, '%')
    FUEL_RAIL_PRESSURE_ABS = C(0x59, 0x02, "FUEL_RAIL_PRESSURE_ABS", "Fuel rail absolute pressure", 0, 655350, "kPa", F("10*(256*A+B)"))
    RELATIVE_ACCELERATOR_POSITION = C(0x5A, 0x01, "RELATIVE_ACCELERATOR_POSITION", "Relative accelerator pedal position", 0, 100, '%', F("100/255*A"))
    HYBRID_BATTERY_REMAINING = C(0x5B, 0x01, "HYBRID_BATTERY_REMAINING", "Hybrid battery pack remaining life", 0, 100, '%', F("100/255*A"))
    OIL_TEMPERATURE = C(0x5C, 0x01, "OIL_TEMPERATURE", "Engine oil temperature", -40, 210, "°C", F("A-40"))
    FUEL_INJECTION_TIMING = C(0x5D, 0x02, "FUEL_INJECTION_TIMING", "Fuel injection timing", -210.00, 301.992, "°", F("(256*A+B)/128-210"))
    FUEL_RATE = C(0x5E, 0x02, "FUEL_RATE", "Engine fuel rate", 0, 3212.75, "L/h", F("(256*A+B)/20"))
    EMISSION_REQUIREMENTS_VEHICLE_DESIGNED = C(0x5F, 0x01, "EMISSION_REQUIREMENTS_VEHICLE_DESIGNED", "Emission requirements to which vehicle is designed", None, None, None)

# PIDS_D = C(0x60, 0x04, "PIDS_D", "PIDs supported [$61 - $80]", None, None, None)
# DRIVER"S_DEMAND_ENGINE_-_PERCENT_TORQUE = C(61, 1, "Driver"s demand engine - percent torque", "Driver"s demand engine - percent torque", -125, 130, '%')
# ACTUAL_ENGINE_-_PERCENT_TORQUE = C(62, 1, "Actual engine - percent torque", "Actual engine - percent torque", -125, 130, '%')
# ENGINE_REFERENCE_TORQUE = C(63, 2, "Engine reference torque", "Engine reference torque", 0, 65535, "N⋅m")
# ENGINE_PERCENT_TORQUE_DATA = C(64, 5, "Engine percent torque data", "Engine percent torque data", -125, 130, '%')
# AUXILIARY_INPUT_/_OUTPUT_SUPPORTED = C(65, 2, "Auxiliary input / output supported", "Auxiliary input / output supported", None, None, None)
# MASS_AIR_FLOW_SENSOR = C(66, 5, "Mass air flow sensor", "Mass air flow sensor", 0, 2047.96875, "g/s")
# ENGINE_COOLANT_TEMPERATURE = C(67, 3, "Engine coolant temperature", "Engine coolant temperature", -40, 215, "°C")
# INTAKE_AIR_TEMPERATURE_SENSOR = C(68, 3, "Intake air temperature sensor", "Intake air temperature sensor", -40, 215, "°C")
# ACTUAL_EGR,_COMMANDED_EGR,_AND_EGR_ERROR = C(69, 7, "Actual EGR, Commanded EGR, and EGR Error", "Actual EGR, Commanded EGR, and EGR Error", None, None, None)
# COMMANDED_DIESEL_INTAKE_AIR_FLOW_CONTROL_AND_RELATIVE_INTAKE_AIR_FLOW_POSITION = C(6A, 5, "Commanded Diesel intake air flow control and relative intake air flow position", "Commanded Diesel intake air flow control and relative intake air flow position", None, None, None)
# EXHAUST_GAS_RECIRCULATION_TEMPERATURE = C(6B, 5, "Exhaust gas recirculation temperature", "Exhaust gas recirculation temperature", None, None, None)
# COMMANDED_THROTTLE_ACTUATOR_CONTROL_AND_RELATIVE_THROTTLE_POSITION = C(6C, 5, "Commanded throttle actuator control and relative throttle position", "Commanded throttle actuator control and relative throttle position", None, None, None)
# FUEL_PRESSURE_CONTROL_SYSTEM = C(6D, 11, "Fuel pressure control system", "Fuel pressure control system", None, None, None)
# INJECTION_PRESSURE_CONTROL_SYSTEM = C(6E, 9, "Injection pressure control system", "Injection pressure control system", None, None, None)
# TURBOCHARGER_COMPRESSOR_INLET_PRESSURE = C(6F, 3, "Turbocharger compressor inlet pressure", "Turbocharger compressor inlet pressure", None, None, None)
# BOOST_PRESSURE_CONTROL = C(70, 10, "Boost pressure control", "Boost pressure control", None, None, None)
# VARIABLE_GEOMETRY_TURBO_(VGT)_CONTROL = C(71, 6, "Variable Geometry turbo (VGT) control", "Variable Geometry turbo (VGT) control", None, None, None)
# WASTEGATE_CONTROL = C(72, 5, "Wastegate control", "Wastegate control", None, None, None)
# EXHAUST_PRESSURE = C(73, 5, "Exhaust pressure", "Exhaust pressure", None, None, None)
# TURBOCHARGER_RPM = C(74, 5, "Turbocharger RPM", "Turbocharger RPM", None, None, None)
# TURBOCHARGER_TEMPERATURE = C(75, 7, "Turbocharger temperature", "Turbocharger temperature", None, None, None)
# TURBOCHARGER_TEMPERATURE = C(76, 7, "Turbocharger temperature", "Turbocharger temperature", None, None, None)
# CHARGE_AIR_COOLER_TEMPERATURE_(CACT) = C(77, 5, "Charge air cooler temperature (CACT)", "Charge air cooler temperature (CACT)", None, None, None)
# EXHAUST_GAS_TEMPERATURE_(EGT)_BANK_1 = C(78, 9, "Exhaust Gas temperature (EGT) Bank 1", "Exhaust Gas temperature (EGT) Bank 1", None, None, None)
# EXHAUST_GAS_TEMPERATURE_(EGT)_BANK_2 = C(79, 9, "Exhaust Gas temperature (EGT) Bank 2", "Exhaust Gas temperature (EGT) Bank 2", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF)_DIFFERENTIAL_PRESSURE = C(7A, 7, "Diesel particulate filter (DPF) differential pressure", "Diesel particulate filter (DPF) differential pressure", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF) = C(7B, 7, "Diesel particulate filter (DPF)", "Diesel particulate filter (DPF)", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF)_TEMPERATURE = C(7C, 9, "Diesel Particulate filter (DPF) temperature", "Diesel Particulate filter (DPF) temperature", None, None, "°C")
# NOX_NTE_(NOT-TO-EXCEED)_CONTROL_AREA_STATUS = C(7D, 1, "NOx NTE (Not-To-Exceed) control area status", "NOx NTE (Not-To-Exceed) control area status", None, None, None)
# PM_NTE_(NOT-TO-EXCEED)_CONTROL_AREA_STATUS = C(7E, 1, "PM NTE (Not-To-Exceed) control area status", "PM NTE (Not-To-Exceed) control area status", None, None, None)
# ENGINE_RUN_TIME_[B] = C(7F, 13, "Engine run time [b]", "Engine run time [b]", None, None, "s")

# PIDS_E = C(80, 4, "PIDs supported [$81 - $A0]", "PIDs supported [$81 - $A0]", None, None, None)
# ENGINE_RUN_TIME_FOR_AUXILIARY_EMISSIONS_CONTROL_DEVICE(AECD) = C(81, 41, "Engine run time for Auxiliary Emissions Control Device(AECD)", "Engine run time for Auxiliary Emissions Control Device(AECD)", None, None, None)
# ENGINE_RUN_TIME_FOR_AUXILIARY_EMISSIONS_CONTROL_DEVICE(AECD) = C(82, 41, "Engine run time for Auxiliary Emissions Control Device(AECD)", "Engine run time for Auxiliary Emissions Control Device(AECD)", None, None, None)
# NOX_SENSOR = C(83, 9, "NOx sensor", "NOx sensor", None, None, None)
# MANIFOLD_SURFACE_TEMPERATURE = C(84, 1, "Manifold surface temperature", "Manifold surface temperature", None, None, None)
# NOX_REAGENT_SYSTEM = C(85, 10, "NOx reagent system", "NOx reagent system", None, None, '%')
# PARTICULATE_MATTER_(PM)_SENSOR = C(86, 5, "Particulate matter (PM) sensor", "Particulate matter (PM) sensor", None, None, None)
# INTAKE_MANIFOLD_ABSOLUTE_PRESSURE = C(87, 5, "Intake manifold absolute pressure", "Intake manifold absolute pressure", None, None, None)
# SCR_INDUCE_SYSTEM = C(88, 13, "SCR Induce System", "SCR Induce System", None, None, None)
# RUN_TIME_FOR_AECD_#11-#15 = C(89, 41, "Run Time for AECD #11-#15", "Run Time for AECD #11-#15", None, None, None)
# RUN_TIME_FOR_AECD_#16-#20 = C(8A, 41, "Run Time for AECD #16-#20", "Run Time for AECD #16-#20", None, None, None)
# DIESEL_AFTERTREATMENT = C(8B, 7, "Diesel Aftertreatment", "Diesel Aftertreatment", None, None, None)
# O2_SENSOR_(WIDE_RANGE) = C(8C, 17, "O2 Sensor (Wide Range)", "O2 Sensor (Wide Range)", None, None, None)
# THROTTLE_POSITION_G = C(8D, 1, "Throttle Position G", "Throttle Position G", 0, 100, '%')
# ENGINE_FRICTION_-_PERCENT_TORQUE = C(8E, 1, "Engine Friction - Percent Torque", "Engine Friction - Percent Torque", -125, 130, '%')
# PM_SENSOR_BANK_1_&_2 = C(8F, 7, "PM Sensor Bank 1 & 2", "PM Sensor Bank 1 & 2", None, None, None)
# WWH-OBD_VEHICLE_OBD_SYSTEM_INFORMATION = C(90, 3, "WWH-OBD Vehicle OBD System Information", "WWH-OBD Vehicle OBD System Information", None, None, "h")
# WWH-OBD_VEHICLE_OBD_SYSTEM_INFORMATION = C(91, 5, "WWH-OBD Vehicle OBD System Information", "WWH-OBD Vehicle OBD System Information", None, None, "h")
# FUEL_SYSTEM_CONTROL = C(92, 2, "Fuel System Control", "Fuel System Control", None, None, None)
# WWH-OBD_VEHICLE_OBD_COUNTERS_SUPPORT = C(93, 3, "WWH-OBD Vehicle OBD Counters support", "WWH-OBD Vehicle OBD Counters support", None, None, "h")
# NOX_WARNING_AND_INDUCEMENT_SYSTEM = C(94, 12, "NOx Warning And Inducement System", "NOx Warning And Inducement System", None, None, None)
# EXHAUST_GAS_TEMPERATURE_SENSOR = C(98, 9, "Exhaust Gas Temperature Sensor", "Exhaust Gas Temperature Sensor", None, None, None)
# EXHAUST_GAS_TEMPERATURE_SENSOR = C(99, 9, "Exhaust Gas Temperature Sensor", "Exhaust Gas Temperature Sensor", None, None, None)
# HYBRID/EV_VEHICLE_SYSTEM_DATA,_BATTERY,_VOLTAGE = C(9A, 6, "Hybrid/EV Vehicle System Data, Battery, Voltage", "Hybrid/EV Vehicle System Data, Battery, Voltage", None, None, None)
# DIESEL_EXHAUST_FLUID_SENSOR_DATA = C(9B, 4, "Diesel Exhaust Fluid Sensor Data", "Diesel Exhaust Fluid Sensor Data", None, None, '%')
# O2_SENSOR_DATA = C(9C, 17, "O2 Sensor Data", "O2 Sensor Data", None, None, None)
# ENGINE_FUEL_RATE = C(9D, 4, "Engine Fuel Rate", "Engine Fuel Rate", None, None, "g/s")
# ENGINE_EXHAUST_FLOW_RATE = C(9E, 2, "Engine Exhaust Flow Rate", "Engine Exhaust Flow Rate", None, None, "kg/h")
# FUEL_SYSTEM_PERCENTAGE_USE = C(9F, 9, "Fuel System Percentage Use", "Fuel System Percentage Use", None, None, None)

# PIDS_F = C(A0, 4, "PIDs supported [$A1 - $C0]", "PIDs supported [$A1 - $C0]", None, None, None)
# NOX_SENSOR_CORRECTED_DATA = C(A1, 9, "NOx Sensor Corrected Data", "NOx Sensor Corrected Data", None, None, "ppm")
# CYLINDER_FUEL_RATE = C(A2, 2, "Cylinder Fuel Rate", "Cylinder Fuel Rate", 0, 2047.96875, "mg/stroke")
# EVAP_SYSTEM_VAPOR_PRESSURE = C(A3, 9, "Evap System Vapor Pressure", "Evap System Vapor Pressure", None, None, "Pa")
# TRANSMISSION_ACTUAL_GEAR = C(A4, 4, "Transmission Actual Gear", "Transmission Actual Gear", 0, 65.535, "ratio")
# COMMANDED_DIESEL_EXHAUST_FLUID_DOSING = C(A5, 4, "Commanded Diesel Exhaust Fluid Dosing", "Commanded Diesel Exhaust Fluid Dosing", 0, 127.5, '%')
# ODOMETER_[C] = C(A6, 4, "Odometer [c]", "Odometer [c]", 0, 429496729.5, "km")
# NOX_SENSOR_CONCENTRATION_SENSORS_3_AND_4 = C(A7, 4, "NOx Sensor Concentration Sensors 3 and 4", "NOx Sensor Concentration Sensors 3 and 4", None, None, None)
# NOX_SENSOR_CORRECTED_CONCENTRATION_SENSORS_3_AND_4 = C(A8, 4, "NOx Sensor Corrected Concentration Sensors 3 and 4", "NOx Sensor Corrected Concentration Sensors 3 and 4", None, None, None)
# ABS_DISABLE_SWITCH_STATE = C(A9, 4, "ABS Disable Switch State", "ABS Disable Switch State", None, None, None)

# PIDS_G = C(C0, 4, "PIDs supported [$C1 - $E0]", "PIDs supported [$C1 - $E0]", None, None, None)
# FUEL_LEVEL_INPUT_A/B = C(C3, 2, "Fuel Level Input A/B", "Fuel Level Input A/B", 0, 25700, '%')
# EXHAUST_PARTICULATE_CONTROL_SYSTEM_DIAGNOSTIC_TIME/COUNT = C(C4, 8, "Exhaust Particulate Control System Diagnostic Time/Count", "Exhaust Particulate Control System Diagnostic Time/Count", 0, 4294967295, "seconds / Count")
# FUEL_PRESSURE_A_AND_B = C(C5, 4, "Fuel Pressure A and B", "Fuel Pressure A and B", 0, 5177, "kPa")
# BYTE_1_-_PARTICULATE_CONTROL_-_DRIVER_INDUCEMENT_SYSTEM_STATUS_BYTE_2,3_-_REMOVAL_OR_BLOCK_OF_THE_PARTICULATE_AFTERTREATMENT_SYSTEM_COUNTER_BYTE_4,5_-_LIQUID_REGENT_INJECTION_SYSTEM_(E.G._FUEL-BORNE_CATALYST)_FAILURE_COUNTER_BYTE_6,7_-_MALFUNCTION_OF_PARTICULATE_CONTROL_MONITORING_SYSTEM_COUNTER = C(C6, 7, "Byte 1 - Particulate control - driver inducement system status Byte 2,3 - Removal or block of the particulate aftertreatment system counter Byte 4,5 - Liquid regent injection system (e.g. fuel-borne catalyst) failure counter Byte 6,7 - Malfunction of Particulate control monitoring system counter", "Byte 1 - Particulate control - driver inducement system status Byte 2,3 - Removal or block of the particulate aftertreatment system counter Byte 4,5 - Liquid regent injection system (e.g. fuel-borne catalyst) failure counter Byte 6,7 - Malfunction of Particulate control monitoring system counter", 0, 65535, "h")
# DISTANCE_SINCE_REFLASH_OR_MODULE_REPLACEMENT = C(C7, 2, "Distance Since Reflash or Module Replacement", "Distance Since Reflash or Module Replacement", 0, 65535, "km")
# NOX_CONTROL_DIAGNOSTIC_(NCD)_AND_PARTICULATE_CONTROL_DIAGNOSTIC_(PCD)_WARNING_LAMP_STATUS = C(C8, 1, "NOx Control Diagnostic (NCD) and Particulate Control Diagnostic (PCD) Warning Lamp status", "NOx Control Diagnostic (NCD) and Particulate Control Diagnostic (PCD) Warning Lamp status", -, -, "Bit")
# DESCRIPTION = C(PID (hex), Data bytes returned, "Description", "Description", Min value, Max value, "Units")


from setuptools import setup
import re


def derive_version() -> str:
    version = ''
    with open('obdii/__init__.py') as f:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

    if not version:
        raise RuntimeError('version is not set')

    if version.endswith(('a', 'b', 'rc')):
        # append version identifier based on commit count
        try:
            import subprocess

            p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, _ = p.communicate()
            if out:
                version += out.decode('utf-8').strip()
            p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, _ = p.communicate()
            if out:
                version += '+g' + out.decode('utf-8').strip()
        except Exception:
            pass

    return version


setup(version=derive_version())


__title__ = "obdii"
__author__ = "PaulMarisOUMary"
__license__ = "MIT"
__copyright__ = "Copyright 2025-present PaulMarisOUMary"
__version__ = "0.1.0a5"

from logging import NullHandler, getLogger

from .connection import Connection
from .commands import commands
from .modes.modeat import at_commands

# We must __init__ .protocols to BaseProtocol.register supported protocols
from .protocols import *

__all__ = [
    "Connection",
    "commands",
    "at_commands",
]

getLogger(__name__).addHandler(NullHandler())